# Cursor movement
HOME = "\x1b[H"  # ]
UP = "\x1b[A"  # ]
DOWN = "\x1b[B"  # ]
RIGHT = "\x1b[C"  # ]
LEFT = "\x1b[D"  # ]
START = "\r"

# Screen clearing
CLEAR_SCREEN_END = "\x1b[0J"  # ]
CLEAR_SCREEN_START = "\x1b[1J"  # ]
CLEAR_SCREEN = "\x1b[2J"  # ]
CLEAR_LINE_END = "\x1b[0K"  # ]
CLEAR_LINE_START = "\x1b[1K"  # ]
CLEAR_LINE = "\x1b[2K"  # ]

# Text styles
BOLD = "\x1b[1m"  # ]
DIM = "\x1b[2m"  # ]
ITALIC = "\x1b[3m"  # ]
UNDERLINE = "\x1b[4m"  # ]
BLINK = "\x1b[5m"  # ]
REVERSE = "\x1b[7m"  # ]
HIDDEN = "\x1b[8m"  # ]
STRIKE = "\x1b[9m"  # ]

# Text colors
RESET = "\x1b[0m"  # ]
BLACK = "\x1b[30m"  # ]
RED = "\x1b[31m"  # ]
GREEN = "\x1b[32m"  # ]
YELLOW = "\x1b[33m"  # ]
BLUE = "\x1b[34m"  # ]
MAGENTA = "\x1b[35m"  # ]
CYAN = "\x1b[36m"  # ]
WHITE = "\x1b[37m"  # ]
BLACK_BRIGHT = "\x1b[90m"  # ]
RED_BRIGHT = "\x1b[91m"  # ]
GREEN_BRIGHT = "\x1b[92m"  # ]
YELLOW_BRIGHT = "\x1b[93m"  # ]
BLUE_BRIGHT = "\x1b[94m"  # ]
MAGENTA_BRIGHT = "\x1b[95m"  # ]
CYAN_BRIGHT = "\x1b[96m"  # ]
WHITE_BRIGHT = "\x1b[97m"  # ]

# Background colors
BG_BLACK = "\x1b[40m"  # ]
BG_RED = "\x1b[41m"  # ]
BG_GREEN = "\x1b[42m"  # ]
BG_YELLOW = "\x1b[43m"  # ]
BG_BLUE = "\x1b[44m"  # ]
BG_MAGENTA = "\x1b[45m"  # ]
BG_CYAN = "\x1b[46m"  # ]
BG_WHITE = "\x1b[47m"  # ]
BG_BLACK_BRIGHT = "\x1b[100m"  # ]
BG_RED_BRIGHT = "\x1b[101m"  # ]
BG_GREEN_BRIGHT = "\x1b[102m"  # ]
BG_YELLOW_BRIGHT = "\x1b[103m"  # ]
BG_BLUE_BRIGHT = "\x1b[104m"  # ]
BG_MAGENTA_BRIGHT = "\x1b[105m"  # ]
BG_CYAN_BRIGHT = "\x1b[106m"  # ]
BG_WHITE_BRIGHT = "\x1b[107m"  # ]



import numpy as np
import time

from ...dataclasses import TrainInfo
from ...log import ansi


def format_time(seconds: int) -> str:
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    parts = []
    if hours:
        parts.append(f"{hours}h")
    if minutes:
        parts.append(f"{minutes}m")
    parts.append(f"{seconds}s")

    return " ".join(parts)


class TrainingLogger:
    def __init__(self, task: str, epochs: int) -> None:
        self.task: str = task
        self.epochs: int = epochs

        self.train_losses: list[float] = []
        self.val_losses: list[float] = []

        self.lr_values: list[float] = []

        self.avg_time_per_epoch: float = 0.0
        self.start_time: float = time.time()

    def log(self,
            train_loss: float,
            val_loss: float,
            lr: float,
            epoch: int,
            time_per_epoch: float,
            ) -> None:
        self.train_losses.append(train_loss)
        self.val_losses.append(val_loss)
        self.lr_values.append(lr)

        self.avg_time_per_epoch = (self.avg_time_per_epoch * (epoch - 1) + time_per_epoch) / epoch

    def get_info(self) -> TrainInfo:
        return TrainInfo(
            train_losses=np.array(self.train_losses),
            best_train_loss=np.min(self.train_losses),
            val_losses=np.array(self.val_losses),
            best_val_loss=np.min(self.val_losses),
            lr_values=np.array(self.lr_values),
        )

    def __repr__(self) -> str:
        # progress
        epoch: int = len(self.train_losses)
        percentage: int = int(epoch / self.epochs * 100)

        # loss
        train_loss: str = f"{self.train_losses[-1]:.4f}"
        val_loss: str = f"{self.val_losses[-1]:.4f}"
        best_val_loss: str = f"{min(self.val_losses):.4f}"

        # time
        time_from_start: int = int(time.time() - self.start_time)
        time_remaining: int = int(self.avg_time_per_epoch * (self.epochs - len(self.train_losses)))

        # message
        clear = ansi.START + (ansi.UP + ansi.CLEAR_LINE)*3
        epoch_msg = f"   |> epoch:   {epoch}/{self.epochs} ({percentage}%)\n"
        loss_msg = f"   |> loss:    train({train_loss}), val({val_loss}), best({best_val_loss})\n"
        time_msg = f"   |> time:    {format_time(time_from_start)} ({format_time(time_remaining)} remaining)"

        return f"{clear if epoch != 1 else ''}{epoch_msg}{loss_msg}{time_msg}"



import optuna
import time

from ...log import ansi


def format_time(seconds: int) -> str:
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    parts = []
    if hours:
        parts.append(f"{hours}h")
    if minutes:
        parts.append(f"{minutes}m")
    parts.append(f"{seconds}s")

    return " ".join(parts)


class OptimizationLogger:
    def __init__(self, n_trials: int) -> None:
        optuna.logging.set_verbosity(optuna.logging.ERROR)

        self.trial_number = 0
        self.best_trial_number = 0
        self.best_metric = 0.0
        self.n_trials = n_trials
        self.avg_time_per_trial = 0.0
        self.start_time = time.time()

    def log(self,
            trial_number: int,
            metric: float,
            best_trial_number: int,
            best_metric: float,
            time_elapsed: float,
            ) -> None:
        self.trial_number = trial_number
        self.metric = metric
        self.best_trial_number = best_trial_number
        self.best_metric = best_metric

        self.avg_time_per_trial = (self.avg_time_per_trial * (trial_number - 1) + time_elapsed) / trial_number

    def __repr__(self) -> str:
        # progress
        percentage = int(self.trial_number / self.n_trials * 100)

        # time
        time_remaining: int = int(self.avg_time_per_trial * (self.n_trials - self.trial_number))
        time_from_start: int = int(time.time() - self.start_time)

        # message
        clear = ansi.START + (ansi.UP + ansi.CLEAR_LINE)*4
        trial_msg = f"   |> trial:           {self.trial_number}/{self.n_trials} ({percentage}%)\n"
        last_metric = f"   |> metric:          {self.metric:.4f}\n"
        best_metric = f"   |> best metric:     {self.best_metric:.4f} (at trial {self.best_trial_number})\n"
        time_msg: str = f"   |> time:            {format_time(time_from_start)} ({format_time(time_remaining)} remaining)"

        return f"{clear if self.trial_number != 1 else ''}{trial_msg}{last_metric}{best_metric}{time_msg}"



from torch.nn import Identity
from catasta import Scaffold, Dataset
from catasta.models import GPHeadRegressor, TransformerRegressor
from catasta.dataclasses import TrainInfo


def single_output() -> None:
    pre_model = TransformerRegressor(
        n_inputs=1,
        n_outputs=1,
        n_patches=1,
        d_model=16,
        n_heads=1,
        n_layers=1,
        feedforward_dim=16,
        head_dim=16,
        dropout=0.1,
    )
    pre_model.linear_head = Identity()  # type: ignore

    model = GPHeadRegressor(
        pre_model=pre_model,
        pre_model_output_dim=16,
        n_inputs=1,
        n_outputs=1,
        n_inducing_points=16,
        kernel="rq",
        mean="constant",
        likelihood="gaussian",
    )

    dataset_root: str = "data/incomplete/"
    dataset = Dataset(
        dataset_root,
        task="regression",
        input_name="input",
        output_name="output",
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
    )

    train_info: TrainInfo = scaffold.train(
        epochs=1000,
        batch_size=128,
        lr=1e-3,
    )
    print(train_info)

    info = scaffold.evaluate()
    print(info)


def multi_output() -> None:
    pre_model = TransformerRegressor(
        n_inputs=4,
        n_outputs=3,
        n_patches=1,
        d_model=16,
        n_heads=1,
        n_layers=1,
        feedforward_dim=16,
        head_dim=16,
        dropout=0.1,
    )
    pre_model.linear_head = Identity()  # type: ignore

    model = GPHeadRegressor(
        pre_model=pre_model,
        pre_model_output_dim=16,
        n_inputs=4,
        n_outputs=3,
        n_inducing_points=16,
        kernel="rq",
        mean="constant",
        likelihood="gaussian",
    )

    # dataset_root: str = "data/tactile/"
    dataset_root: str = "data/squishy/"
    dataset = Dataset(
        dataset_root,
        task="regression",
        input_name=["s0", "s1", "s2", "s3"],
        output_name=["x", "y", "fz"],
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
    )

    train_info: TrainInfo = scaffold.train(
        epochs=5000,
        batch_size=128,
        lr=1e-3,
    )
    print(train_info)

    info = scaffold.evaluate()
    print(info)


if __name__ == '__main__':
    # single_output()
    multi_output()



from torch.nn import Sequential, Linear
from einops.layers.torch import Rearrange, Reduce

from catasta import Scaffold, Dataset
from catasta.models import GPHeadRegressor
from catasta.dataclasses import TrainInfo


def single_output() -> None:
    n_inputs: int = 1
    n_patches: int = 1
    patch_size: int = n_inputs // n_patches
    d_model: int = 16
    pooling: str = "concat"
    pre_model_output_dim = d_model * n_patches if pooling == "concat" else d_model

    pre_model = Sequential(
        Rearrange('b (n p) -> b n p', p=patch_size),
        Linear(patch_size, d_model),
        Rearrange('b n d -> b (n d)') if pooling == "concat" else Reduce('b n d -> b d', pooling)
    )

    model = GPHeadRegressor(
        pre_model=pre_model,
        pre_model_output_dim=pre_model_output_dim,
        n_inputs=1,
        n_outputs=1,
        n_inducing_points=16,
        kernel="rq",
        mean="constant",
        likelihood="gaussian",
    )

    dataset_root: str = "data/incomplete/"
    dataset = Dataset(
        dataset_root,
        task="regression",
        input_name="input",
        output_name="output",
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
    )

    train_info: TrainInfo = scaffold.train(
        epochs=1000,
        batch_size=128,
        lr=1e-3,
    )
    print(train_info)

    info = scaffold.evaluate()
    print(info)


def multi_output() -> None:
    n_inputs: int = 4
    n_patches: int = 1
    patch_size: int = n_inputs // n_patches
    d_model: int = 16
    pooling: str = "concat"
    pre_model_output_dim = d_model * n_patches if pooling == "concat" else d_model

    pre_model = Sequential(
        Rearrange('b (n p) -> b n p', p=patch_size),
        Linear(patch_size, d_model),
        Rearrange('b n d -> b (n d)') if pooling == "concat" else Reduce('b n d -> b d', pooling)
    )

    model = GPHeadRegressor(
        pre_model=pre_model,
        pre_model_output_dim=pre_model_output_dim,
        n_inputs=4,
        n_outputs=3,
        n_inducing_points=16,
        kernel="rq",
        mean="constant",
        likelihood="gaussian",
    )

    # dataset_root: str = "data/tactile/"
    dataset_root: str = "data/squishy/"
    dataset = Dataset(
        dataset_root,
        task="regression",
        input_name=["s0", "s1", "s2", "s3"],
        output_name=["x", "y", "fz"],
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
    )

    train_info: TrainInfo = scaffold.train(
        epochs=5000,
        batch_size=128,
        lr=1e-3,
    )
    print(train_info)

    info = scaffold.evaluate()
    print(info)


if __name__ == '__main__':
    single_output()
    multi_output()



from catasta import Scaffold, Dataset
from catasta.models import CNNImageClassifier
from catasta.dataclasses import EvalInfo


def main() -> None:
    model = CNNImageClassifier(
        input_shape=(28, 28, 3),
        n_classes=10,
        conv_out_channels=[16, 32],
        feedforward_dims=[16, 16],
    )

    dataset = Dataset(
        "data/reduced_mnist",
        task="classification",
    )

    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="cross_entropy",
    )

    scaffold.train(
        epochs=100,
        batch_size=128,
        lr=1e-3,
    )

    eval_info: EvalInfo = scaffold.evaluate(batch_size=32)
    print(eval_info)


if __name__ == '__main__':
    main()
